# ExpansionTree.py,v 1.26 2003/08/13 03:16:55 jamie v1.1
__version__ = "1.26"
__all__ = ()

import zLOG
from Tree import Tree, TreeMaker
from binascii import b2a_base64, a2b_base64, Error
from operator import isMappingType
from string import maketrans

# The object id encoding/decoding routines. (we only use the encoder)
_b2a_map = maketrans("+/=", "-._")
_a2b_map = maketrans("-._", "+/=")
__all__ += ("b2a", "a2b")
def b2a(b): return b2a_base64(str(b))[:-1].translate(_b2a_map)
def a2b(a): return a2b_base64(a.translate(_a2b_map))


class _UseDefaultState(Exception):
    """internal Exception raised to change logic flow during decoding"""

__all__ += ("LimitExceeded",)
class LimitExceeded(Exception):
    """Exception raised when a saftey limit is exceeded"""


__all__ += ("ExpansionTree",)
class ExpansionTree(Tree):
    __allow_access_to_unprotected_subobjects__ = 1

    def shape(self):
        """Returns encoded instructions for changing the state of this node.

        If this node is a branch, returns a string which can be used with the
        `ExpansionTreeMaker.grow()` method to modify the given tree state,
        otherwise if this node is a leaf returns `None`.
        """
        state = self.state
        if state == 0: return None
        if state > 0: return "c,%d,%s" % (self.expansion_number, self.id)
        return "e,%d,%s" % (self.aq_parent.expansion_number, self.id)


__all__ += ("ExpansionTreeMaker",)
class ExpansionTreeMaker(TreeMaker):
    __allow_access_to_unprotected_subobjects__ = 1

    def __init__(self, limit, function):
        """Extends `TreeMaker`'s constructor.

        'limit' -- *integer*, when set to a non-negative value no more than
        `limit` separators may appear in the `state` string given to
        `ExpansionTreeMaker.grow()`.  This must be used to prevent denial of
        service attacks when handling state data from untrusted sources.  A
        tree with *n* branch nodes can have a maximum of 2*n* - 2 separators
        in its encoded state string.

        'function' -- see `TreeMaker.__init__()`
        """
        TreeMaker.__init__(self, function)
        self.limit = int(limit)

    def encode(self, tree):
        """Returns a string encoding of `tree`, safe for use in cookies,
        queries, URLs, or wherever.  `tree` is assumed to be a preorder list
        of nodes, like that produced by `Tree.flatten()`.
        """
        depth = -1
        eids = []
        n = 0
        for node in tree:
            if node.state <= 0: continue
            delta = depth - node.depth + 1
            depth = node.depth
            if delta > 0: eids.append('_' * delta)
            eids.append(node.id)
            node.expansion_number = n
            n += 1
        return ':'.join(eids)

    def decode(self, s, limit, nth=None):
        """Returns a decoded state structure from string `s`.  If `nth` is an
        integer, return the (state, nid) pair for the nth entry as well.
        """
        state = p = {} # TreeMaker state structure & position reference
        pstack = []
        pop = 0
        nth_pair = nth is not None and (None, None) or None
        if nth < 0: nth = None
        try:
            if limit >= 0 and s.count(':') > limit: raise LimitExceeded
            for eid in s.split(':'):
                if eid.startswith('_'):
                    pop = len(eid) - 1
                    continue
                if len(eid) < 4: raise Error # b2a() returns atleast 4 bytes
                if pop < 0:
                    pstack.append(p) # record our position
                    p[nid] = {}      # make a new branch
                    p = p[nid]       # point to the new branch
                elif state:
                    p[nid] = None
                nid = eid 
                if pop > 0:
                    p = pstack[-pop]  # jump back up the tree
                    del pstack[-pop:] # forget about that branch
                pop = -1
                if nth == 0:
                    nth_pair = (p, nid)
                    nth = None
                elif nth is not None:
                    nth -= 1
            p[nid] = None
            if nth == 0: return state, (p, nid)
        except (UnboundLocalError, IndexError, Error):
            # pstack operations may raise IndexError if eid depth is bogus
            # final p[nid] assignment may raise UnboundLocalError if goaded
            # we raise Error if we see any bogus eid segments
            # note: we don't catch LimitExceeded, the caller should
            return nth_pair is not None and (None, (None, None)) or None
        return nth_pair is not None and (state, nth_pair) or state

    def node(self, object, b2a=b2a):
        """Overrides `TreeMaker.node()`.

        Returns an `ExpansionTree` with instance attributes 'object'
        populated by a reference to `object`, 'id' populated by the `b2a`
        encoded results of `getId()`, and 'state' set to 0.
        """
        node = ExpansionTree()
        node.state = 0
        node.object = object
        node.id = b2a(self.getId(object))
        return node

    def grow(self, root_object, state, shape, default_state={}):
        """Generates and returns the tuple of data used to generate the markup
        of a tree widget rooted at `root_object` and expanded according to
        `state` as modified by the instructions in `shape`.

        The tuple returned contains three objects: the root `ExpansionTree`,
        the entire tree as a flattened preorder list of `ExpansionTree`s,
        and a string, safe for use in URLs, queries, and cookies, that is an
        encoded representation of the state. 

        Arguments are as follows:

        'state' -- An encoded state *string* of the current tree.

        'shape' -- *String* of encoded instructions for modifying the current
        tree, as generated by the 'ExpansionTree.shape()' method.  An empty
        string implies no modifications.

        'default_state' -- Set this to change the default state of your tree.
        If you pass in a *mapping* it will be treated as the unencoded tree 
        state as used by 'TreeMaker.make()', this is largely for programatic
        use.  If you pass in a *string* it will be treated as an encoded state
        string, just like the 'state' argument, this is the human-friendly way
        to set the default state.
        """
        try:
            if not state: raise _UseDefaultState
            try:
                v, n, target_eid = shape.split(',')
                state, (p, eid) = self.decode(state, self.limit, int(n))
                if p is None: pass
                elif v == 'e':
                    if p[eid] is None:
                        p[eid] = {target_eid:None}
                    else:
                        p[eid][target_eid] = None
                elif v == 'c' and p is not state and eid == target_eid:
                    del p[eid]
            except ValueError: # invalid shape
                state = self.decode(state, self.limit)
            if state is None: raise _UseDefaultState
        except (LimitExceeded, _UseDefaultState), e:
            if isinstance(e, LimitExceeded):
                zLOG.LOG("ExpansionTree", zLOG.WARNING,
                         "expansion limit exceeded, using default state")
            if isinstance(default_state, str):
                state = self.decode(default_state, self.limit)
            elif isMappingType(default_state):
                state = default_state
            else:
                raise TypeError, "default_state not StringType or MappingType"
        tree = self.make(root_object, state)
        rows = tree.flatten()
        return tree, rows, self.encode(rows)
