# Tree.py,v 1.15 2003/11/01 22:49:03 jamie v1.1
# Adapted from ZTUtils/Tree.py,v 1.6 2002/08/14 22:10:12 mj Exp
# Copyright 2001 Zope Corporation and Contributors. All Rights Reserved.
__version__ = "1.15"

from Acquisition import Explicit
from ComputedAttribute import ComputedAttribute
from operator import isMappingType


class Tree(Explicit):
    """A simple tree class built on top of acquisition.

    'depth' -- The depth of this node; if the root node has a depth of 0, its
    immediate children will have a depth of 1, and so forth.  As depth is a
    `ComputedAttribute` the root note must have its depth assigned, thus
    allowing its child nodes to compute their depth accordingly.

    'height' -- The height of the tree beneath this node; a leaf has a height
    of 1, its parent a height of 2, and so forth.

    'size' -- The number of nodes in the tree.
    """
    depth = ComputedAttribute(lambda s: s.aq_parent.depth + 1, 1)
    height = 1
    size = 1

    def __init__(self):
        self._child_list = []

    def __getitem__(self, index):
        return self._child_list[index].__of__(self)

    def __len__(self):
        return len(self._child_list)

    def _walk(self, f, data=None):
        if data is None:
            f(self)
        else:
            f(self, data)
        for child in self._child_list:
            child.__of__(self)._walk(f, data)

    def graft(self, child):
        """Adds a child `Tree` to this node."""
        self._child_list.append(child)
        self.height = max(self.height, child.height + 1)
        self.size = self.size + child.size

    def flatten(self):
        """Returns a flattened, preordered, list of `Tree` nodes."""
        items = []
        self._walk(items.append)
        return items


class TreeMaker:
    """Utility for mapping a hierachy of objects into a `Tree` instance.

    Tree Anatomy & Terminology::

          A        Branch nodes are shown as capital letters.
         /|\       Leaf nodes are shown as lower case letters.
        B m C      A is the root node, or trunk.
       /   / \     Nodes A, B, C, E, and F are expanded.
      D-  E   F    Node D is collapsed.
         / \  |
        n   o p
    """
    def __init__(self, function):
        """Constructs and returns a TreeMaker instance.

        Starting at the root object, `function` is used on every object in the
        tree to determine that object's children.  `function` must take the
        object to test as its first argument and return a list of child objects.
        """
        self._child_function = function

    def getChildren(self, object):
        """Returns a list of objects determined to be the children of `object`.

        Child determination is done using the user-supplied function
        established during instantiation of a `TreeMaker`.
        """
        return self._child_function(object)

    def hasChildren(self, object):
        """Returns True if `object` has children, False if it does not.

        A call to `getChildren()` will be performed to establish the existence
        of any children.
        """
        return bool(self.getChildren(object))

    def getId(self, object):
        """Returns `object`'s unique id."""
        return hasattr(object, "_p_oid") and str(object._p_oid) or id(object)

    def node(self, object):
        """Returns a `Tree` with instance attributes 'object' populated by a
        reference to `object`, and 'id' populated by the results of `getId()`,
        and 'state' set to 0.
        """
        node = Tree()
        node.state = 0 # leaf
        node.object = object
        node.id = self.getId(object)
        return node

    def make(self, root, expansion=None, subtree=False):
        """Returns a `Tree` grown from `root` pruned according to `expansion`.
        """
        node = self.node(root)
        if not subtree: node.depth = 0
        child_xpn = expansion
        if isMappingType(expansion):
            expansion = expansion.has_key(node.id)
            child_xpn = child_xpn.get(node.id)
        if expansion or not subtree:
            children = self.getChildren(root)
            if children:
                node.state = 1 # expanded branch
                for child in children:
                    node.graft(self.make(child, child_xpn, True))
        elif self.hasChildren(root): node.state = -1 # collapsed branch
        return node
